from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit_aer import AerSimulator
import numpy as np

class QKD:
    def __init__(self, message, share_size):
        self.message = message
        self.n = len(message)
        self.random_filter_order()
        self.share_size = share_size

    def random_filter_order (self):
        self.filter_order = np.random.randint(2, size=self.n) + 1

    def decrypt_message (self):
        simulator = AerSimulator()        
        string = []
        for i in range(self.n):
            qc = self.message[i]
            if self.filter_order[i] == 1:
                qc.measure(0, 0)
                results = simulator.run(qc, shots=1).result().get_counts()
                result = sorted(results, key=results.get, reverse=True)[0]
                string.append(result)
            else:
                qc.h(0)
                qc.measure(0,0)
                results = simulator.run(qc, shots=1).result().get_counts()
                result = sorted(results, key=results.get, reverse=True)[0]
                string.append(result)
        self.string = string

    def share_filter_order (self):
        return self.filter_order

    def remove_bits (self, remove_indices):
        self.string = np.delete(self.string, remove_indices)
        self.string = [str(i) for i in self.string]
        self.string = ''.join(self.string)

    def share_secret_string (self):
        return self.string[:self.share_size]


class PersonA (QKD):
    def __init__ (self, message, share_size):
        QKD.__init__(self, message, share_size)
    
    def is_spy (self, answer):
        if answer:
            print("INCORRECT!")
        else:
            print("CORRECT!")


class PersonB (QKD):
    def __init__ (self, message, share_size):
        QKD.__init__(self, message, share_size)

    def spy_interception (self):
        # The spy just measures all of the transmitted circuits
        # to determine their value
        for m in self.message:
            m.measure(0,0)
            simulator = AerSimulator()
            results = simulator.run(m, shots=1).result().get_counts()
            # Sort the returned dictionary in order of counts of each result
            result = sorted(results, key=results.get, reverse=True)[0]

    def decrypt_message (self):
        simulator = AerSimulator()
        self.spy_interception()
        string = []
        simulator = AerSimulator()

        for i in range(self.n):
            qc = self.message[i]
            if self.filter_order[i] == 1:
                qc.measure(0, 0)
                results = simulator.run(qc, shots=1).result().get_counts()
                result = sorted(results, key=results.get, reverse=True)[0]
                string.append(result)
            else:
                qc.h(0)
                qc.measure(0,0)
                results = simulator.run(qc, shots=1).result().get_counts()
                result = sorted(results, key=results.get, reverse=True)[0]
                string.append(result)
        self.string = string

    def is_spy (self, answer):
        if answer:
            print("CORRECT!")
        else:
            print("INCORRECT!")

            
